import os
import re
from functools import reduce
from .data import WARP
from .helpers import (
    cleanName,
    isClean,
    specFromRanges,
    rangesFromList,
    setFromSpec,
    nbytes,
)

# If a feature, with type string, has less than ENUM_LIMIT values,
# an enumeration type for it will be created
# provided all values of that feature are a valid name for MQL.

ENUM_LIMIT = 1000

ONE_ENUM_TYPE = True


class MQL(object):
    def __init__(self, mqlDir, mqlName, tfFeatures, tm):
        self.mqlDir = mqlDir
        cleanDb = cleanName(mqlName)
        if cleanDb != mqlName:
            self.tm.error('db name "{}" => "{}"'.format(mqlName, cleanDb))
        self.mqlName = cleanDb
        self.tfFeatures = tfFeatures
        self.tm = tm
        self.enums = {}
        self._check()

    def write(self):
        if not self.good:
            return
        if not os.path.exists(self.mqlDir):
            try:
                os.makedirs(self.mqlDir, exist_ok=True)
            except Exception:
                self.tm.error(
                    'Cannot create directory "{}"'.format(self.mqlDir)
                )
                self.good = False
                return
        mqlPath = '{}/{}.mql'.format(self.mqlDir, self.mqlName)
        try:
            fm = open(mqlPath, 'w', encoding='utf8')
        except Exception:
            self.tm.error('Could not write to {}'.format(mqlPath))
            self.good = False
            return

        self.tm.info('Loading {} features'.format(len(self.featureList)))
        for ft in self.featureList:
            fObj = self.features[ft]
            fObj.load()

        self.fm = fm
        self._writeStartDb()
        self._writeEnums()
        self._writeTypes()
        self._writeDataAll()
        self._writeEndDb()
        self.tm.indent(level=0)
        self.tm.info('Done')

    def _check(self):
        self.tm.info('Checking features of dataset {}'.format(self.mqlName))
        self.features = {}
        self.featureList = []
        self.tm.indent(level=1)
        for (f, fo) in sorted(self.tfFeatures.items()):
            if fo.method is not None or f in WARP:
                continue
            fo.load(metaOnly=True)
            if fo.isConfig:
                continue
            cleanF = cleanName(f)
            if cleanF != f:
                self.tm.error('feature "{}" => "{}"'.format(f, cleanF))
            self.featureList.append(cleanF)
            self.features[cleanF] = fo
        good = True
        for feat in (WARP[0], WARP[1], '__levels__'):
            if feat not in self.tfFeatures:
                self.tm.error(
                    '{} feature {} is missing from data set'.format(
                        'Warp' if feat in WARP else 'Computed'
                        if feat.startswith('__') else 'Data',
                        feat,
                    )
                )
                good = False
            else:
                fObj = self.tfFeatures[feat]
                if not fObj.load():
                    good = False
        self.tm.indent(level=0)
        if (not good):
            self.tm.error('Export to MQL aborted')
        else:
            self.tm.info(
                '{} features to export to MQL ...'.format(
                    len(self.featureList)
                )
            )
        self.good = good

    def _writeStartDb(self):
        self.fm.write(
            '''
CREATE DATABASE '{name}'
GO
USE DATABASE '{name}'
GO
'''.format(name=self.mqlName)
        )

    def _writeEndDb(self):
        self.fm.write('''
VACUUM DATABASE ANALYZE
GO
''')
        self.fm.close()

    def _writeEnums(self):
        self.tm.indent(level=0)
        self.tm.info('Writing enumerations')
        self.tm.indent(level=1)
        for ft in self.featureList:
            fObj = self.features[ft]
            if fObj.isEdge or fObj.dataType == 'int':
                continue
            fMap = fObj.data
            fValues = sorted(set(fMap.values()))
            if len(fValues) > ENUM_LIMIT:
                continue
            eligible = all(isClean(fVal) for fVal in fValues)
            if not eligible:
                unclean = [fVal for fVal in fValues if not isClean(fVal)]
                print(
                    '\t{:<15}: {:>4} values, {} not a name, e.g. «{}»'.format(
                        ft,
                        len(fValues),
                        len(unclean),
                        unclean[0],
                    )
                )
                continue
            self.enums[ft] = fValues

        if ONE_ENUM_TYPE:
            self._writeEnumsAsOne()
        else:
            for ft in sorted(self.enums):
                self._writeEnum(ft)
            self.tm.indent(level=0)
            self.tm.info('Written {} enumerations'.format(len(self.enums)))

    def _writeEnumsAsOne(self):
        fValues = reduce(
            set.union,
            (set(fV) for fV in self.enums.values()),
            set(),
        )
        self.tm.info(
            'Writing an all-in-one enum with {:>4} values'.format(
                len(fValues)
            )
        )
        fValuesEnumerated = ',\n\t'.join(
            '{} = {}'.format(fVal, i) for (i, fVal) in enumerate(fValues)
        )
        self.fm.write(
            '''
CREATE ENUMERATION all_enum = {{
    {}
}}
GO
'''.format(fValuesEnumerated)
        )

    def _writeEnum(self, ft):
        fValues = self.enums[ft]
        self.tm.info('enum {:<15} with {:>4} values'.format(ft, len(fValues)))
        fValuesEnumerated = ',\n\t'.join(
            '{} = {}'.format(fVal, i) for (i, fVal) in enumerate(fValues)
        )
        self.fm.write(
            '''
CREATE ENUMERATION {}_enum = {{
    {}
}}
GO
'''.format(ft, fValuesEnumerated)
        )

    def _writeTypes(self):
        def valInt(n):
            return str(n)

        def valStr(s):
            if "'" in s:
                return '"{}"'.format(s.replace('"', '\\"'))
            else:
                return "'{}'".format(s)

        def valIds(ids):
            return '({})'.format(','.join(str(i) for i in ids))

        self.levels = self.tfFeatures['__levels__'].data[::-1]
        self.tm.indent(level=0)
        self.tm.info(
            'Mapping {} features onto {} object types'.format(
                len(self.featureList),
                len(self.levels),
            )
        )
        otypeSupport = {}
        for (otype, av, start, end) in self.levels:
            cleanOtype = cleanName(otype)
            if cleanOtype != otype:
                self.tm.error('otype "{}" => "{}"'.format(otype, cleanOtype))
            otypeSupport[cleanOtype] = set(range(start, end + 1))

        self.otypes = {}
        self.featureTypes = {}
        self.featureMethods = {}
        for ft in self.featureList:
            fObj = self.features[ft]
            if fObj.isEdge:
                dataType = 'LIST OF id_d'
                method = valIds
            else:
                if fObj.dataType == 'str':
                    dataType = 'string DEFAULT ""'
                    method = valInt if ft in self.enums else valStr
                elif fObj.dataType == 'int':
                    dataType = 'integer DEFAULT 0'
                    method = valInt
                else:
                    dataType = 'string DEFAULT ""'
                    method = valStr
            self.featureTypes[ft] = dataType
            self.featureMethods[ft] = method

            support = set(fObj.data.keys())
            for otype in otypeSupport:
                if len(support & otypeSupport[otype]):
                    self.otypes.setdefault(otype, []).append(ft)

        for otype in (cleanName(x[0]) for x in self.levels):
            self._writeType(otype)

    def _writeType(self, otype):
        self.fm.write('''
CREATE OBJECT TYPE
[{}
'''.format(otype))
        for ft in self.otypes[otype]:
            fType = '{}_enum'.format(
                'all' if ONE_ENUM_TYPE else ft
            ) if ft in self.enums else self.featureTypes[ft]
            self.fm.write('  {}:{};\n'.format(ft, fType))
        self.fm.write('''
]
GO
''')

    def _writeDataAll(self):
        self.tm.info(
            'Writing {} features as data in {} object types'.format(
                len(self.featureList),
                len(self.levels),
            )
        )
        self.oslots = self.tfFeatures[WARP[1]].data
        for (otype, av, start, end) in self.levels:
            self._writeData(otype, start, end)

    def _writeData(self, otype, start, end):
        tm = self.tm
        fm = self.fm
        tm.indent(level=1, reset=True)
        tm.info('{} data ...'.format(otype))
        oslots = self.oslots
        maxSlot = oslots[-1]
        oFeats = self.otypes[otype]
        features = self.features
        featureMethods = self.featureMethods
        fm.write(
            '''
DROP INDEXES ON OBJECT TYPE[{o}]
GO
CREATE OBJECTS
WITH OBJECT TYPE[{o}]
'''.format(o=otype)
        )
        curSize = 0
        LIMIT = 50000
        t = 0
        j = 0
        tm.indent(level=2, reset=True)
        for n in range(start, end + 1):
            oMql = '''
CREATE OBJECT
FROM MONADS= {{ {m} }}
WITH ID_D={i} [
'''.format(
                m=n if n <= maxSlot else
                specFromRanges(rangesFromList(oslots[n - maxSlot - 1])),
                i=n,
            )
            for ft in oFeats:
                method = featureMethods[ft]
                fMap = features[ft].data
                if n in fMap:
                    oMql += '{}:={};\n'.format(ft, method(fMap[n]))
            oMql += '''
]
'''
            fm.write(oMql)
            curSize += len(bytes(oMql, encoding='utf8'))
            t += 1
            j += 1
            if j == LIMIT:
                fm.write(
                    '''
GO
CREATE OBJECTS
WITH OBJECT TYPE[{o}]
'''.format(o=otype)
                )
                tm.info(
                    'batch of size {:>20} with {:>7} of {:>7} {}s'.format(
                        nbytes(curSize), j, t, otype
                    )
                )
                j = 0
                curSize = 0

        tm.info(
            'batch of size {:>20} with {:>7} of {:>7} {}s'.format(
                nbytes(curSize), j, t, otype
            )
        )
        fm.write(
            '''
GO
CREATE INDEXES ON OBJECT TYPE[{o}]
GO
'''.format(o=otype)
        )

        tm.indent(level=1)
        tm.info('{} data: {} objects'.format(otype, t))


# MQL IMPORT

uniscan = re.compile(r'(?:\\x..)+')


def makeuni(match):
    ''' Make proper unicode of a text that contains byte escape codes
        such as backslash xb6
    '''
    byts = eval('"' + match.group(0) + '"')
    return byts.encode('latin1').decode('utf-8')


def uni(line):
    return uniscan.sub(makeuni, line)


def tfFromMql(mqlFile, tm, slotType=None, otext=None, meta=None):
    if slotType is None:
        tm.error('ERROR: no slotType specified')
        return (False, {}, {}, {})
    (good, objectTypes, tables, edgeF, nodeF) = parseMql(mqlFile, tm)
    if not good:
        return (False, {}, {}, {})
    return tfFromData(
        tm, objectTypes, tables, edgeF, nodeF, slotType, otext, meta
    )


def parseMql(mqlFile, tm):
    tm.info('Parsing mql source ...')
    fh = open(mqlFile)

    objectTypes = dict()
    tables = dict()

    edgeF = dict()
    nodeF = dict()

    curId = None
    curEnum = None
    curObjectType = None
    curTable = None
    curObject = None
    curValue = None
    curFeature = None
    seeObjects = False

    inObjectTypeFeatures = False

    STRING_TYPES = {'ascii', 'string'}

    enums = dict()

    chunkSize = 1000000
    inThisChunk = 0

    good = True

    for (ln, line) in enumerate(fh):
        inThisChunk += 1
        if inThisChunk == chunkSize:
            tm.info('\tline {:>9}'.format(ln + 1))
            inThisChunk = 0
        if line.startswith('CREATE OBJECTS WITH OBJECT TYPE'
                           ) or line.startswith('WITH OBJECT TYPE'):
            comps = line.rstrip().rstrip(']').split('[', 1)
            curTable = comps[1]
            tm.info('\t\tobjects in {}'.format(curTable))
            curObject = None
            if curTable not in tables:
                tables[curTable] = dict()
            seeObjects = True
        elif line == 'CREATE OBJECT\n':
            curObject = None
            curObject = dict(feats=dict(), monads=None)
            curId = None
            seeObjects = True
        elif curEnum is not None:
            if line.startswith('}'):
                curEnum = None
                continue
            comps = line.strip().rstrip(',').split('=', 1)
            comp = comps[0].strip()
            words = comp.split()
            if words[0] == 'DEFAULT':
                enums[curEnum]['default'] = uni(words[1])
                value = words[1]
            else:
                value = words[0]
            enums[curEnum]['values'].append(value)
        elif curObjectType is not None:
            if line.startswith(']'):
                curObjectType = None
                inObjectTypeFeatures = False
                continue
            if curObjectType is True:
                if line.startswith('['):
                    curObjectType = line.rstrip()[1:]
                    objectTypes[curObjectType] = dict()
                    tm.info('\t\totype {}'.format(curObjectType))
                    inObjectTypeFeatures = True
                    continue
            if inObjectTypeFeatures:
                comps = line.strip().rstrip(';').split(':', 1)
                feature = comps[0].strip()
                fInfo = comps[1].strip()
                fCleanInfo = fInfo.replace('FROM SET', '')
                fInfoComps = fCleanInfo.split(' ', 1)
                fMQLType = fInfoComps[0]
                if len(fInfoComps) == 2:
                    fDefaultComps = fInfoComps[1].strip().split(' ', 1)
                    fDefault = fDefaultComps[
                        1
                    ] if len(fDefaultComps) > 1 else None
                else:
                    fDefault = None
                if fDefault is not None and fMQLType in STRING_TYPES:
                    fDefault = uni(fDefault[1:-1])
                default = enums.get(fMQLType, {}).get('default', fDefault)
                ftype = 'str' if fMQLType in enums else\
                        'int' if fMQLType == 'integer' else\
                        'str' if fMQLType in STRING_TYPES else\
                        'int' if fInfo == 'id_d' else\
                        'str'
                isEdge = fMQLType == 'id_d'
                if isEdge:
                    edgeF.setdefault(curObjectType, set()).add(feature)
                else:
                    nodeF.setdefault(curObjectType, set()).add(feature)

                objectTypes[curObjectType][feature] = (ftype, default)
                tm.info(
                    '\t\t\tfeature {} ({}) =def= {} : {}'.format(
                        feature, ftype, default, 'edge' if isEdge else 'node'
                    )
                )
        elif seeObjects:
            if curObject is not None:
                if line.startswith(']'):
                    objectType = objectTypes[curTable]
                    for (feature, (ftype, default)) in objectType.items():
                        if feature not in curObject['feats'
                                                    ] and default is not None:
                            curObject['feats'][feature] = default
                    tables[curTable][curId] = curObject
                    curObject = None
                    continue
                elif line.startswith('['):
                    name = line.rstrip()[1:]
                    if len(name):
                        curTable = name
                        if curTable not in tables:
                            tables[curTable] = dict()
                elif line.startswith('FROM MONADS'):
                    monads = line.split('=', 1)[1].replace('{', '').replace(
                        '}', ''
                    ).replace(' ', '').strip()
                    curObject['monads'] = setFromSpec(monads)
                elif line.startswith('WITH ID_D'):
                    comps = line.replace('[', '').rstrip().split('=', 1)
                    curId = int(comps[1])
                elif line.startswith('GO'):
                    pass
                elif line.strip() == '':
                    pass
                else:
                    if curValue is not None:
                        toBeContinued = not line.rstrip().endswith('";')
                        if toBeContinued:
                            curValue += line
                        else:
                            curValue += line.rstrip().rstrip(';').rstrip('"')
                            curObject['feats'][curFeature] = uni(curValue)
                            curValue = None
                            curFeature = None
                        continue
                    if ':=' in line:
                        (featurePart, valuePart) = line.split('=', 1)
                        feature = featurePart[0:-1].strip()
                        valuePart = valuePart.lstrip()
                        isText = ':="' in line
                        toBeContinued = isText and not line.rstrip(
                        ).endswith('";')
                        if toBeContinued:
                            # this happens if a feature value
                            # contains a new line
                            # we must continue scanning lines
                            # until we meet the end of the value
                            curFeature = feature
                            curValue = valuePart.lstrip('"')
                        else:
                            value = valuePart.rstrip().rstrip(';').strip('"')
                            curObject['feats'][feature] = uni(
                                value
                            ) if isText else value
                    else:
                        tm.error(
                            'ERROR: line {}: unrecognized line -->{}<--'.
                            format(ln, line)
                        )
                        good = False
                        break
            else:
                if line.startswith('CREATE OBJECT'):
                    curObject = dict(feats=dict(), monads=None)
                    curId = None
        else:
            if line.startswith('CREATE ENUMERATION'):
                words = line.split()
                curEnum = words[2]
                enums[curEnum] = dict(default=None, values=[])
                tm.info('\t\tenum {}'.format(curEnum))
            elif line.startswith('CREATE OBJECT TYPE'):
                curObjectType = True
                inObjectTypeFeatures = False
    tm.info('{} lines parsed'.format(ln + 1))
    fh.close()
    for table in tables:
        tm.info('{} objects of type {}'.format(len(tables[table]), table))

    if len(tables) == 0:
        tm.info('No objects found')
    return (good, objectTypes, tables, nodeF, edgeF)


def tfFromData(tm, objectTypes, tables, nodeF, edgeF, slotType, otext, meta):
    tm.info('Making TF data ...')

    NIL = {'nil', 'NIL', 'Nil'}

    tableOrder = [slotType] + [t for t in sorted(tables) if t != slotType]

    iddFromMonad = dict()
    slotFromMonad = dict()

    nodeFromIdd = dict()
    iddFromNode = dict()

    nodeFeatures = dict()
    edgeFeatures = dict()
    metaData = dict()

    # metadata that ends up in every feature
    metaData[''] = {} if meta is None else meta

    # the config feature otext
    metaData['otext'] = otext

    good = True

    tm.info('Monad - idd mapping ...')
    for idd in tables.get(slotType, {}):
        monad = list(tables[slotType][idd]['monads'])[0]
        iddFromMonad[monad] = idd

    tm.info('Removing holes in the monad sequence')
    # we set up a monad - slot mapping
    curSlot = 0
    otype = dict()
    for monad in sorted(iddFromMonad):
        curSlot += 1
        slotFromMonad[monad] = curSlot
        idd = iddFromMonad[monad]
        nodeFromIdd[idd] = curSlot
        iddFromNode[curSlot] = idd
        otype[curSlot] = slotType

    maxSlot = curSlot
    tm.info('maxSlot={}'.format(maxSlot))

    tm.info('Node mapping and otype ...')
    node = maxSlot
    for t in tableOrder[1:]:
        for idd in sorted(tables[t]):
            node += 1
            nodeFromIdd[idd] = node
            iddFromNode[node] = idd
            otype[node] = t

    nodeFeatures['otype'] = otype
    metaData['otype'] = dict(valueType='str', )

    tm.info('oslots ...')
    oslots = dict()
    for t in tableOrder[1:]:
        for idd in tables.get(t, {}):
            node = nodeFromIdd[idd]
            monads = tables[t][idd]['monads']
            oslots[node] = {slotFromMonad[m] for m in monads}
    edgeFeatures['oslots'] = oslots
    metaData['oslots'] = dict(valueType='str', )

    tm.info('metadata ...')
    for t in nodeF:
        for f in nodeF[t]:
            ftype = objectTypes[t][f][0]
            metaData.setdefault(f, {})['valueType'] = ftype
    for t in edgeF:
        for f in edgeF[t]:
            metaData.setdefault(f, {})['valueType'] = 'str'

    tm.info('features ...')
    chunkSize = 100000
    for t in tableOrder:
        tm.info('\tfeatures from {}s'.format(t))
        inThisChunk = 0
        thisTable = tables.get(t, {})
        for (i, idd) in enumerate(thisTable):
            inThisChunk += 1
            if inThisChunk == chunkSize:
                tm.info('\t{:>9} {}s'.format(i + 1, t))
                inThisChunk = 0
            node = nodeFromIdd[idd]
            features = tables[t][idd]['feats']
            for (f, v) in features.items():
                isEdge = f in edgeF.get(t, set())
                if isEdge:
                    if v not in NIL:
                        edgeFeatures.setdefault(f, {}).setdefault(
                            node, set()
                        ).add(nodeFromIdd[int(v)])
                else:
                    nodeFeatures.setdefault(f, {})[node] = v
        tm.info('\t{:>9} {}s'.format(len(thisTable), t))

    return (good, nodeFeatures, edgeFeatures, metaData)



import re


class Transcription(object):
    decomp = {
        '\u05E9\u05C1': "\uFB2A",
        '\u05E9\u05C2': "\uFB2B",
    }
    hebrew_mapping = {
        '_': ' ',  # space inside word
        '92': "\u0591",  # etnahta = atnach
        '01': "\u0592",  # segolta
        '65': "\u0593",  # shalshelet
        '80': "\u0594",  # zaqef_qatan
        '85': "\u0595",  # zaqef_gadol
        '73': "\u0596",  # tipeha = tifcha
        '81': "\u0597",  # revia = rebia
        '82': "\u0598",  # zarqa = tsinorit = zinorit = sinnorit
        '03': "\u0599",  # pashta
        '10': "\u059A",  # yetiv = yetib
        '91': "\u059B",  # tevir = tebir
        '61': "\u059C",  # geresh
        '11': "\u059D",  # geresh muqdam = mugrash
        '62': "\u059E",  # gershayim = garshayim
        '84': "\u059F",  # qarney para = pazer_gadol
        '14': "\u05A0",  # telisha_gedola
        '44': "\u05A0",  # telisha_gedola = telisha_gedola_med
        '83': "\u05A1",  # pazer
        '74': "\u05A3",  # munah = munach
        '70': "\u05A4",  # mahapakh = mehuppach
        '71': "\u05A5",  # merkha = merecha
        '72': "\u05A6",  # merkha kefula = merecha_kepula
        '94': "\u05A7",  # darga
        '63': "\u05A8",  # qadma = azla
        '33': "\u05A8",  # pashta_med < qadma
        '04': "\u05A9",  # telisha_qetana
        '24': "\u05A9",  # telisha_qetana = telisha_qetana_med
        '93': "\u05AA",  # yera ben yomo = yerach
        '60': "\u05AB",  # ole = ole_weyored
        '64': "\u05AC",  # iluy = illuy
        '13': "\u05AD",  # dehi = dechi
        '02': "\u05AE",  # zinor = sinnor
        '*': "\u05AF",  # masora circle
        ':': "\u05B0",  # sheva = shewa
        ':E': "\u05B1",  # hataf segol = chataph_segol
        ':A': "\u05B2",  # hataf patah = chataph_patach
        ':@': "\u05B3",  # hataf qamats = chataph_qamats
        'I': "\u05B4",  # hiriq = chiriq
        ';': "\u05B5",  # tsere
        'E': "\u05B6",  # segol
        'A': "\u05B7",  # patach
        '@': "\u05B8",  # qamats
        'O': "\u05B9",  # holam = cholam
        'U': "\u05BB",  # qubuts = qubbuts
        '.': "\u05BC",  # dagesh
        '25': "\u05BD",  # silluq yamin
        '45': "\u05BD",  # meteg
        '35': "\u05BD",  # meteg (tikon)
        '75': "\u05BD",  # siluq = silluq
        '95': "\u05BD",  # meteg = meteg_yamin
        '&': "\u05BE",  # maqaf
        ',': "\u05BF",  # rafe = raphe
        '05': "\u05C0",  # paseq
        '.c': "\u05C1",  # shin dot
        '.f': "\u05C2",  # sin dot
        '00': "\u05C3",  # sof_pasuq
        '52': "\u05C4",  # upper dot = puncta_above
        '53': "\u05C5",  # lower dot = puncta_below
        'ñ': "\u05C6\u0307",  # nun hafukha
        'Ñ': "\u05C6\u0307",  # nun hafukha
        '>': "\u05D0",  # alef
        'B': "\u05D1",  # bet
        'G': "\u05D2",  # gimel
        'D': "\u05D3",  # dalet
        'H': "\u05D4",  # he
        'W': "\u05D5",  # vav
        'Z': "\u05D6",  # zayin
        'X': "\u05D7",  # het
        'V': "\u05D8",  # tet
        'J': "\u05D9",  # yod
        'k': "\u05DA",  # kaf final
        'K': "\u05DB",  # kaf
        'L': "\u05DC",  # lamed
        'm': "\u05DD",  # mem final
        'M': "\u05DE",  # mem
        'n': "\u05DF",  # nun final
        'N': "\u05E0",  # nun
        'S': "\u05E1",  # samekh
        '<': "\u05E2",  # ayin
        'p': "\u05E3",  # pe final
        'P': "\u05E4",  # pe
        'y': "\u05E5",  # tsadi final
        'Y': "\u05E6",  # tsadi
        'Q': "\u05E7",  # qof
        'R': "\u05E8",  # resh
        '#': "\u05E9",  # sin unpointed
        'T': "\u05EA",  # tav
        'C': "\uFB2A",  # shin pointed
        'F': "\uFB2B",  # sin pointed
        '55': "<UNMAPPED 55=large letter>",  # large_letter
        '56': "<UNMAPPED 56=small letter>",  # small_letter
        '57': "<UNMAPPED 57=suspended letter>",  # suspended_letter
        '-': "",  # suppress space afterward
    }
    hebrew_cons = '>BGDHWZXVJKLMNS<PYQRFCT'
    trans_final_pat = re.compile(
        r'([' + hebrew_cons + r'][^_&]*)([KMNPY])([^' + hebrew_cons +
        r'_&]*(:?[_&]|\Z))'
    )
    trans_hebrew_pat = re.compile(r'(:[AE@]|.[cf]|:|[0-9][0-9]|.)')
    swap_accent_pat = re.compile(
        r'(\A|[_&])([0-9][0-9])([' + hebrew_cons + r'])([:;,.EAIOU@]*)'
    )
    remove_accent_pat = re.compile(r'((?:[0-9][0-9])|[,*])')
    remove_point_pat = re.compile(
        r'((?:[0-9][0-9])|(?:\.[cf])|(?::[@AE])|[,.:;@AEIOU*])'
    )
    remove_psn_pat = re.compile(r'00[ _SPNÑñ]*')
    remove_psq_pat = re.compile(r'(?:[ _]+05[ _]*)|(?:05[ _]+)')
    shin_pat = re.compile(r'[CF]')
    ph_simple_pat = re.compile(r'([ˈˌᵊᵃᵒᵉāo*])')
    noorigspace = re.compile(
        '''
          (?: [&-]\Z)           # space, maqef or nospace
        | (?:
               0[05]            # sof pasuq or paseq
               (?:_[SNP])*      # nun hafukha, setumah, petuhah at end of verse
               \Z
          )
        | (?:_[SPN])+           #  nun hafukha, setumah, petuhah between words
    ''', re.X
    )

    syriac_mapping = {
        '>': "\u0710",  # alaph
        'B': "\u0712",  # beth
        'G': "\u0713",  # gamal
        'D': "\u0715",  # dalat
        'H': "\u0717",  # he
        'W': "\u0718",  # waw
        'Z': "\u0719",  # zain
        'X': "\u071A",  # heth
        'V': "\u071B",  # teth
        'J': "\u071D",  # yudh
        'K': "\u071F",  # kaph
        'L': "\u0720",  # lamadh
        'M': "\u0721",  # mim
        'N': "\u0722",  # nun
        'S': "\u0723",  # semkath
        '<': "\u0725",  # e
        'P': "\u0726",  # pe
        'Y': "\u0728",  # sadhe
        'Q': "\u0729",  # qaph
        'R': "\u072A",  # rish
        'C': "\u072B",  # shin
        'T': "\u072C",  # taw
        's': "\u0724",  # semkath final
        'p': "\u0727",  # pe reversed
    }

    def __init__(self):
        self.hebrew_consonants = {
            Transcription.hebrew_mapping[x]
            for x in Transcription.hebrew_cons
        }
        self.hebrew_consonants.add('\u05E9')
        self.hebrew_mappingi = dict(
            (v, k) for (k, v) in Transcription.hebrew_mapping.items()
            if k != ''
        )
        # special treatment needed for nun hafukha,
        # since it is consists of two characters
        self.hebrew_mappingi['\u05C6'] = 'ñ'
        self.hebrew_mappingi['\u0307'] = ''

    def _comp(s):
        for (d, c) in Transcription.decomp.items():
            s = s.replace(d, c)
        return s

    def _decomp(s):
        for (d, c) in Transcription.decomp.items():
            s = s.replace(c, d)
        return s

    def suffix_and_finales(word):
        # first split the word proper from the suffix,
        # and add a space if there is no other suffix
        add_space = ''
        suffix = ''
        new_word = word
        if not word:
            return (new_word, suffix + add_space)
        lastch = new_word[-1]
        if lastch == '-' or lastch == '&':
            new_word = new_word[0:-1]
            suffix = lastch
        else:
            if len(new_word) >= 2:
                lastch = new_word[-1]
                llastch = new_word[-2]
                if llastch == '_' and (lastch == 'P' or lastch == 'S'):
                    new_word = new_word[0:-2]
                    suffix = ' ' + lastch + suffix + ' '
            if len(new_word) >= 2:
                lastch = new_word[-1]
                llastch = new_word[-2]
                if llastch == '_' and (lastch == 'N'):
                    new_word = new_word[0:-2]
                    suffix = ' ñ' + suffix + ' '
            if len(new_word) >= 2:
                lastch = new_word[-1]
                llastch = new_word[-2]
                if llastch == '0' and (lastch == '0' or lastch == '5'):
                    new_word = new_word[0:-2]
                    suffix = (
                        ' ' if lastch == '5' else ''
                    ) + llastch + lastch + suffix
                    add_space = '\n' if lastch == '0' else ' '
        if suffix == '':
            add_space = ' '
        elif suffix == '-':
            add_space = ''
            suffix = ''
        # second: replace consonants by their final forms when needed
        new_word = Transcription.trans_final_pat.sub(
            Transcription._map_final, new_word
        )
        return (new_word, suffix + add_space)

    def _map_final(m):
        return m.group(1) + m.group(2).lower() + m.group(3)

    def _map_hebrew(m):
        return Transcription.hebrew_mapping.get(m.group(1), m.group(1))

    def _swap_accent(m):
        return m.group(1) + m.group(3) + m.group(4) + m.group(2)

    def _remove_accent(m):
        return '00' if m.group(1) == '00' else '05' if m.group(
            1
        ) == '05' else ''

    def _remove_point(m):
        return '00' if m.group(1) == '00' else '05' if m.group(
            1
        ) == '05' else ''

    def _ph_simple(m):
        return 'å' if m.group(1) in 'āo' else ''

    # unicode normalization is harmful
    # if there is a combination of dagesh, vowel and accent.

    def suppress_space(word):
        return Transcription.noorigspace.search(word)

    def to_etcbc_v(word):
        return Transcription.remove_accent_pat.sub(
            Transcription._remove_accent, word
        )

    def to_etcbc_c(word):
        word = Transcription.remove_point_pat.sub(
            Transcription._remove_point, word
        )
        word = Transcription.remove_psn_pat.sub(
            '00', word
        )  # remove nun hafukha, setumah, petuhah at the end of a verse
        word = Transcription.remove_psq_pat.sub(
            ' ', word
        )  # replace paseq with attached spaces by single space
        word = word.upper()  # no final forms of consonants
        return Transcription.shin_pat.sub('#', word)

    def to_hebrew(word):
        word = Transcription.swap_accent_pat.sub(
            Transcription._swap_accent, word
        )
        return Transcription.trans_hebrew_pat.sub(
            Transcription._map_hebrew, word
        )

    def to_hebrew_v(word):
        return Transcription.trans_hebrew_pat.sub(
            Transcription._map_hebrew, Transcription.to_etcbc_v(word)
        )

    def to_hebrew_c(word):
        return Transcription.trans_hebrew_pat.sub(
            Transcription._map_hebrew, Transcription.to_etcbc_c(word)
        )

    def to_hebrew_x(word):
        return Transcription.trans_hebrew_pat.sub(
            Transcription._map_hebrew, word
        )

    def ph_simplify(pword):
        return Transcription.ph_simple_pat.sub(Transcription._ph_simple, pword)

    def from_hebrew(self, word):
        return ''.join(
            self.hebrew_mappingi.get(x, x) for x in Transcription._comp(word)
        )

    def to_syriac(self, word):
        return Transcription._decomp(
            ''.join(
                self.syriac_mapping.get(x, x)
                for x in Transcription._comp(word)
            )
        )

    def from_syriac(self, word):
        return ''.join(
            self.syriac_mappingi.get(x, x) for x in Transcription._comp(word)
        )



import re

LETTER = set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ')
VALID = set('_0123456789') | LETTER


def cleanName(name):
    clean = ''.join(c if c in VALID else '_' for c in name)
    if clean == '' or not clean[0] in LETTER:
        clean = 'x' + clean
    return clean


def isClean(name):
    if name is None or len(name) == 0 or name[0] not in LETTER:
        return False
    return all(c in VALID for c in name[1:])


def expandDir(dirName, paths):
    if dirName.startswith('~'):
        dirName = dirName.replace('~', paths['home'], 1)
    elif dirName.startswith('..'):
        dirName = dirName.replace('..', paths['up'], 1)
    elif dirName.startswith('.'):
        dirName = dirName.replace('.', paths['cur'], 1)
    return dirName


def setFromSpec(spec):
    covered = set()
    for r_str in spec.split(','):
        bounds = r_str.split('-')
        if len(bounds) == 1:
            covered.add(int(r_str))
        else:
            b = int(bounds[0])
            e = int(bounds[1])
            if (e < b):
                (b, e) = (e, b)
            for n in range(b, e + 1):
                covered.add(n)
    return covered


def rangesFromSet(nodeSet):
    ranges = []
    curstart = None
    curend = None
    for n in sorted(nodeSet):
        if curstart is None:
            curstart = n
            curend = n
        elif n == curend + 1:
            curend = n
        else:
            ranges.append((curstart, curend))
            curstart = n
            curend = n
    if curstart is not None:
        ranges.append((curstart, curend))
    return ranges


def rangesFromList(nodeList):  # the list must be sorted
    ranges = []
    curstart = None
    curend = None
    for n in nodeList:
        if curstart is None:
            curstart = n
            curend = n
        elif n == curend + 1:
            curend = n
        else:
            ranges.append((curstart, curend))
            curstart = n
            curend = n
    if curstart is not None:
        ranges.append((curstart, curend))
    return ranges


def specFromRanges(ranges):  # ranges must be normalized
    return ','.join(
        '{}'.format(r[0]) if r[0] == r[1] else '{}-{}'.format(*r)
        for r in ranges
    )


def valueFromTf(tf):
    return '\\'.join(
        x.replace('\\t', '\t').replace('\\n', '\n') for x in tf.split('\\\\')
    )


def tfFromValue(val):
    return str(val) if type(val) is int else val.replace('\\', '\\\\').replace(
        '\t', '\\t'
    ).replace('\n', '\\n')


def makeInverse(data):
    inverse = {}
    for n in data:
        for m in data[n]:
            inverse.setdefault(m, set()).add(n)
    return inverse


def makeInverseVal(data):
    inverse = {}
    for n in data:
        for (m, val) in data[n].items():
            inverse.setdefault(m, {})[n] = val
    return inverse


def nbytes(by):
    units = ['B', 'KB', 'MB', 'GB', 'TB']
    for i in range(len(units)):
        if by < 1024 or i == len(units) - 1:
            fmt = '{:>5}{}' if i == 0 else '{:>5.1f}{}'
            return fmt.format(by, units[i])
        by /= 1024


def collectFormats(config):
    varPattern = re.compile(r'\{([^}]+)\}')
    featureSet = set()

    def collectFormat(tpl):
        features = []

        def varReplace(match):
            varText = match.group(1)
            fts = tuple(varText.split('/'))
            features.append(fts)
            for ft in fts:
                featureSet.add(ft)
            return '{}'

        rtpl = varPattern.sub(varReplace, tpl)
        return (rtpl, tuple(features))

    formats = {}
    for (fmt, tpl) in sorted(config.items()):
        if fmt.startswith('fmt:'):
            formats[fmt[4:]] = collectFormat(tpl)
    return (formats, sorted(featureSet))


def compileFormats(cformats, features):
    xformats = {}
    for (fmt, (rtpl, feats)) in sorted(cformats.items()):
        tpl = rtpl.replace('\\n', '\n').replace('\\t', '\t')
        xformats[fmt] = compileFormat(tpl, feats, features)
    return xformats


def compileFormat(rtpl, feats, features):
    replaceFuncs = []
    for feat in feats:
        replaceFuncs.append(makeFunc(feat, features))

    def g(n):
        values = tuple(replaceFunc(n) for replaceFunc in replaceFuncs)
        return rtpl.format(*values)

    return g


def makeFunc(feat, features):
    if len(feat) == 1:
        ft = feat[0]
        f = features[ft].data
        return (lambda n: f.get(n, ''))
    elif len(feat) == 2:
        (ft1, ft2) = feat
        f1 = features[ft1].data
        f2 = features[ft2].data
        return (lambda n: (f1.get(n, f2.get(n, ''))))


def itemize(string, sep=None):
    if not string:
        return []
    if not sep:
        return string.strip().split()
    return string.strip().split(sep)



import array
import collections
import functools
from .helpers import itemize


def getOtypeInfo(info, otype):
    result = (otype[-2], otype[-1], len(otype) - 2 + otype[-1])
    info('slot={}:1-{};node-{}'.format(*result))
    return result


def levels(info, error, otype, oslots, otext):
    (slotType, maxSlot, maxNode) = getOtypeInfo(info, otype)
    levelOrder = otext.get('levels', None)
    if levelOrder is not None:
        levelRank = {
            level: i for (i, level) in enumerate(levelOrder.split(','))
        }
    otypeCount = collections.Counter()
    otypeMin = {}
    otypeMax = {}
    slotSetLengths = collections.Counter()
    info('get ranking of otypes')
    for k in range(len(oslots) - 1):
        ntp = otype[k]
        otypeCount[ntp] += 1
        slotSetLengths[ntp] += len(oslots[k])
        tfn = k + maxSlot + 1
        if ntp not in otypeMin:
            otypeMin[ntp] = tfn
        if ntp not in otypeMax or otypeMax[ntp] < tfn:
            otypeMax[ntp] = tfn
    sortKey = (
        (lambda x: -x[1]) if levelOrder is None else
        (lambda x: levelRank[x[0]])
    )
    result = tuple(
        sorted(
            ((
                ntp, slotSetLengths[ntp] / otypeCount[ntp], otypeMin[ntp],
                otypeMax[ntp]
            ) for ntp in otypeCount),
            key=sortKey,
        ) + [(slotType, 1, 1, maxSlot)]
    )
    info('results:')
    for (otp, av, omin, omax) in result:
        info(
            '{:<15}: {:>8} {{{}-{}}}'.format(otp, round(av, 2), omin, omax),
            tm=False
        )
    return result


def order(info, error, otype, oslots, levels):
    (slotType, maxSlot, maxNode) = getOtypeInfo(info, otype)
    info('assigning otype levels to nodes')
    otypeLevels = dict(((x[0], i) for (i, x) in enumerate(levels)))

    def otypeRank(n):
        return otypeLevels[slotType
                           if n < maxSlot + 1 else otype[n - maxSlot - 1]]

    def before(na, nb):
        if na < maxSlot + 1:
            a = na
            sa = {a}
        else:
            a = na - maxSlot
            sa = set(oslots[a - 1])
        if nb < maxSlot + 1:
            b = nb
            sb = {b}
        else:
            b = nb - maxSlot
            sb = set(oslots[b - 1])
        oa = otypeRank(na)
        ob = otypeRank(nb)
        if sa == sb:
            return 0 if oa == ob else -1 if oa < ob else 1
        if sa > sb:
            return -1
        if sa < sb:
            return 1
        am = min(sa - sb)
        bm = min(sb - sa)
        return -1 if am < bm else 1 if bm < am else None

    canonKey = functools.cmp_to_key(before)
    info('sorting nodes')
    nodes = sorted(range(1, maxNode + 1), key=canonKey)
    return array.array('I', nodes)


def rank(info, error, otype, order):
    (slotType, maxSlot, maxNode) = getOtypeInfo(info, otype)
    info('ranking nodes')
    nodesRank = dict(((n, i) for (i, n) in enumerate(order)))
    return array.array('I', (nodesRank[n] for n in range(1, maxNode + 1)))


def levUp(info, error, otype, oslots, rank):
    (slotType, maxSlot, maxNode) = getOtypeInfo(info, otype)
    info('making inverse of edge feature oslots')
    oslotsInv = {}
    for (k, mList) in enumerate(oslots[0:-1]):
        for m in mList:
            oslotsInv.setdefault(m, set()).add(k + 1 + maxSlot)
    info('listing embedders of all nodes')
    embedders = []
    for n in range(1, maxSlot + 1):
        contentEmbedders = oslotsInv[n]
        embedders.append(
            tuple(
                sorted(
                    [
                        m
                        for m in contentEmbedders
                        if m != n
                        # if rank[m - 1] < rank[n - 1]
                    ],
                    key=lambda k: -rank[k - 1],
                )
            )
        )
    for n in range(maxSlot + 1, maxNode + 1):
        mList = oslots[n - maxSlot - 1]
        if len(mList) == 0:
            embedders.append(tuple())
        else:
            contentEmbedders = functools.reduce(
                lambda x, y: x & oslotsInv[y],
                mList[1:],
                oslotsInv[mList[0]],
            )
            embedders.append(
                tuple(
                    sorted(
                        [
                            m
                            for m in contentEmbedders
                            if m != n
                            # if rank[m - 1] < rank[n - 1]
                        ],
                        key=lambda k: -rank[k - 1],
                    )
                )
            )
    return tuple(embedders)


def levDown(info, error, otype, levUp, rank):
    (slotType, maxSlot, maxNode) = getOtypeInfo(info, otype)
    info('inverting embedders')
    inverse = {}
    for n in range(maxSlot + 1, maxNode + 1):
        for m in levUp[n - 1]:
            inverse.setdefault(m, set()).add(n)
    info('turning embeddees into list')
    embeddees = []
    for n in range(maxSlot + 1, maxNode + 1):
        embeddees.append(
            tuple(sorted(
                inverse.get(n, []),
                key=lambda m: rank[m - 1],
            ))
        )
    return tuple(embeddees)


def boundary(info, error, otype, oslots, rank):
    firstSlotsD = {}
    lastSlotsD = {}
    (slotType, maxSlot, maxNode) = getOtypeInfo(info, otype)
    for (k, mList) in enumerate(oslots[0:-1]):
        firstSlotsD.setdefault(mList[0], []).append(k + 1 + maxSlot)
        lastSlotsD.setdefault(mList[-1], []).append(k + 1 + maxSlot)
    firstSlots = []
    lastSlots = []
    for n in range(1, maxSlot + 1):
        firstSlots.append(
            tuple(sorted(firstSlotsD.get(n, []), key=lambda k: -rank[k - 1]))
        )
        lastSlots.append(
            tuple(sorted(lastSlotsD.get(n, []), key=lambda k: rank[k - 1]))
        )
    return (tuple(firstSlots), tuple(lastSlots))


def sections(info, error, otype, oslots, otext, levUp, levels, *sFeats):
    (slotType, maxSlot, maxNode) = getOtypeInfo(info, otype)
    support = dict(((o[0], (o[2], o[3])) for o in levels))
    sTypes = itemize(otext['sectionTypes'], ',')
    sec1 = {}
    sec2 = {}
    c1 = 0
    c2 = 0
    support2 = support[sTypes[2]]
    for n2 in range(support2[0], support2[1] + 1):
        # otherwise we miss the last one
        # for n2 in range(*support[sTypes[2]]):
        n0 = tuple(
            x for x in levUp[n2 - 1] if otype[x - maxSlot - 1] == sTypes[0]
        )[0]
        n1 = tuple(
            x for x in levUp[n2 - 1] if otype[x - maxSlot - 1] == sTypes[1]
        )[0]
        n1s = sFeats[1][n1]
        n2s = sFeats[2][n2]
        if n0 not in sec1:
            sec1[n0] = {}
        if n1s not in sec1[n0]:
            sec1[n0][n1s] = n1
            c1 += 1
        sec2.setdefault(n0, {}).setdefault(n1s, {})[n2s] = n2
        c2 += 1
    info('{} {}s and {} {}s indexed'.format(c1, sTypes[1], c2, sTypes[2]))
    return (sec1, sec2)



import os
import re
import operator
from glob import glob
from shutil import copyfile
from functools import reduce
from IPython.display import display, Markdown, HTML

from tf.fabric import Fabric

SOURCE = 'uruk'
SOURCE_FULL = 'Uruk IV-III'
VERSION = '1.0'
CORPUS = f'tf/{SOURCE}/{VERSION}'
CORPUS_FULL = f'{SOURCE_FULL} (v{VERSION})'
SOURCE_DIR = 'sources/cdli'
IMAGE_DIR = f'{SOURCE_DIR}/images'
TEMP_DIR = '_temp'
REPORT_DIR = 'reports'

LIMIT_SHOW = 100
LIMIT_TABLE = 2000

PHOTO_TO = '{}/tablets/photos'
PHOTO_EXT = 'jpg'

TABLET_TO = '{}/tablets/lineart'
IDEO_TO = '{}/ideographs/lineart'
LINEART_EXT = 'jpg'

LOCAL_DIR = 'cdli-imagery'

URL_GH = 'https://github.com'
URL_NB = 'http://nbviewer.jupyter.org/github'

URL_FORMAT = dict(
    tablet=dict(
        photo=f'https://cdli.ucla.edu/dl/photo/{{}}_d.{PHOTO_EXT}',
        lineart=f'https://cdli.ucla.edu/dl/lineart/{{}}_l.{LINEART_EXT}',
        main=(
            'https://cdli.ucla.edu/search/search_results.php?'
            'SearchMode=Text&ObjectID={}'
        ),
    ),
    ideograph=dict(
        lineart=(
            'https://cdli.ucla.edu/tools/SignLists'
            f'/protocuneiform/archsigns/{{}}.{LINEART_EXT}'
        ),
        main=(
            'https://cdli.ucla.edu/tools/SignLists'
            '/protocuneiform/archsigns.html'
        ),
    ),
)

FLAGS = (
    ('damage', '#'),
    ('remarkable', '!'),
    ('written', ('!(', ')')),
    ('uncertain', '?'),
)

OUTER_QUAD_TYPES = {'sign', 'quad'}

CLUSTER_BEGIN = {'[': ']', '<': '>', '(': ')'}
CLUSTER_END = {y: x for (x, y) in CLUSTER_BEGIN.items()}
CLUSTER_KIND = {'[': 'uncertain', '(': 'properName', '<': 'supplied'}
CLUSTER_BRACKETS = dict((name, (bOpen, CLUSTER_BEGIN[bOpen]))
                        for (bOpen, name) in CLUSTER_KIND.items())

FLEX_STYLE = (
    'display: flex;'
    'flex-flow: row nowrap;'
    'justify-content: flex-start;'
    'align-items: center;'
    'align-content: flex-start;'
)
CAPTION_STYLE = dict(
    top=(
        'display: flex;'
        'flex-flow: column-reverse nowrap;'
        'justify-content: space-between;'
        'align-items: center;'
        'align-content: space-between;'
    ),
    bottom=(
        'display: flex;'
        'flex-flow: column nowrap;'
        'justify-content: space-between;'
        'align-items: center;'
        'align-content: space-between;'
    ),
    left=(
        'display: flex;'
        'flex-flow: row-reverse nowrap;'
        'justify-content: space-between;'
        'align-items: center;'
        'align-content: space-between;'
    ),
    right=(
        'display: flex;'
        'flex-flow: row nowrap;'
        'justify-content: space-between;'
        'align-items: center;'
        'align-content: space-between;'
    ),
)

ITEM_STYLE = ('padding: 0.5rem;')

SIZING = {'height', 'width'}

COMMENT_TYPES = set(
    '''
    tablet
    face
    column
    line
    case
'''.strip().split()
)

CLUSTER_TYPES = dict(
    uncertain='?',
    properName='=',
    supplied='&gt;',
)

ATF_TYPES = set('''
    sign
    quad
    cluster
'''.strip().split())

CSS = '''
<style>
.pnum {
    font-family: sans-serif;
    font-size: small;
    font-weight: bold;
    color: #444444;
}
.nd {
    font-family: monospace;
    font-size: x-small;
    color: #999999;
}
.meta {
    display: flex;
    justify-content: flex-start;
    align-items: flex-start;
    align-content: flex-start;
    flex-flow: row nowrap;
}
.features,.comments {
    display: flex;
    justify-content: flex-start;
    align-items: flex-start;
    align-content: flex-start;
    flex-flow: column nowrap;
}
.children {
    display: flex;
    justify-content: flex-start;
    align-items: flex-start;
    align-content: flex-start;
    border: 0;
    background-color: #ffffff;
}
.children.tablet,.children.face {
    flex-flow: row nowrap;
}
.children.column {
    align-items: stretch;
    flex-flow: column nowrap;
}
.children.line,.children.case {
    align-items: stretch;
    flex-flow: row nowrap;
}
.children.caseh,.children.trminalh {
    flex-flow: row nowrap;
}
.children.casev,.children.trminalv {
    flex-flow: column nowrap;
}
.children.trminal {
    flex-flow: row nowrap;
}
.children.cluster {
    flex-flow: row wrap;
}
.children.quad {
    flex-flow: row wrap;
}
.children.sign {
    flex-flow: column nowrap;
}
.contnr {
    width: fit-content;
}
.contnr.tablet,.contnr.face,.contnr.column,
.contnr.line,.contnr.case,.contnr.trminal,
.contnr.comment,
.contnr.cluster,
.contnr.quad,.contnr.sign {
    display: flex;
    justify-content: flex-start;
    align-items: flex-start;
    align-content: flex-start;
    flex-flow: column nowrap;
    background: #ffffff none repeat scroll 0 0;
    padding:  0.5em 0.1em 0.1em 0.1em;
    margin: 0.8em 0.1em 0.1em 0.1em;
    border-radius: 0.2em;
    border-style: solid;
    border-width: 0.2em;
    font-size: small;
}
.contnr.tablet,.contnr.face,.contnr.column {
    border-color: #bb8800;
}
.contnr.line,.contnr.case,.contnr.trminal {
    border-color: #0088bb;
}
.contnr.cluster {
    flex-flow: row wrap;
    border: 0;
}
.contnr.sign,.contnr.quad {
    border-color: #bbbbbb;
}
.contnr.comment {
    background-color: #ffddaa;
    border: 0.2em solid #eecc99;
    border-radius: 0.3em;
}
.contnr.hl {
    background-color: #ffee66;
}
.lbl.tablet,.lbl.face,.lbl.column,
.lbl.line,.lbl.case,.lbl.trminal,
.lbl.comment,
.lbl.cluster,
.lbl.quad,.lbl.sign {
    margin-top: -1.2em;
    margin-left: 1em;
    background: #ffffff none repeat scroll 0 0;
    padding: 0 0.3em;
    border-style: solid;
    font-size: small;
    color: #0000cc;
    display: block;
}
.lbl.tablet,.lbl.face,.lbl.column {
    border-color: #bb8800;
    border-width: 0.2em;
    border-radius: 0.2em;
    color: #bb8800;
}
.lbl.line,.lbl.case,.lbl.trminal {
    border-color: #0088bb;
    border-width: 0.2em;
    border-radius: 0.2em;
    color: #0088bb;
}
.lbl.comment {
    border-color: #eecc99;
    border-width: 0.2em;
    border-radius: 0.2em;
    color: #eecc99;
    background-color: #ffddaa none repeat scroll 0 0;
}
.lbl.clusterB,.lbl.clusterE {
    padding:  0.5em 0.1em 0.1em 0.1em;
    margin: 0.8em 0.1em 0.1em 0.1em;
    color: #888844;
    font-size: x-small;
}
.lbl.clusterB {
    border-left: 0.3em solid #cccc99;
    border-right: 0;
    border-top: 0;
    border-bottom: 0;
    border-radius: 1em;
}
.lbl.clusterE {
    border-left: 0;
    border-right: 0.3em solid #cccc99;
    border-top: 0;
    border-bottom: 0;
    border-radius: 1em;
}
.lbl.quad,.lbl.sign {
    border-color: #bbbbbb;
    border-width: 0.1em;
    border-radius: 0.1em;
    color: #bbbbbb;
}
.op {
    padding:  0.5em 0.1em 0.1em 0.1em;
    margin: 0.8em 0.1em 0.1em 0.1em;
    font-family: monospace;
    font-size: x-large;
    font-weight: bold;
}
.name {
    font-family: monospace;
    font-size: medium;
    color: #0000bb;
}
.period {
    font-family: monospace;
    font-size: medium;
    font-weight: bold;
    color: #0000bb;
}
.excavation {
    font-family: monospace;
    font-size: medium;
    font-style: italic;
    color: #779900;
}
.text {
    font-family: sans-serif;
    font-size: x-small;
    color: #000000;
}
.srcLn {
    font-family: monospace;
    font-size: medium;
    color: #000000;
}
.srcLnNum {
    font-family: monospace;
    font-size: x-small;
    color: #0000bb;
}
</style>
'''


def dm(md):
    display(Markdown(md))


def _outLink(text, href, title=None):
    titleAtt = '' if title is None else f' title="{title}"'
    return f'<a target="_blank" href="{href}"{titleAtt}>{text}</a>'


def _wrapLink(piece, objectType, kind, identifier, pos='bottom', caption=None):
    title = (
        'to CDLI main page for this item'
        if kind == 'main' else f'to higher resolution {kind} on CDLI'
    )
    url = URL_FORMAT.get(objectType, {}).get(kind, '').format(identifier)

    result = _outLink(piece, url, title=title) if url else piece
    if caption:
        result = (
            f'<div style="{CAPTION_STYLE[pos]}">'
            f'<div>{result}</div>'
            f'<div>{caption}</div>'
            '</div>'
        )
    return result


class Cunei(object):
    def __init__(self, repoBase, repoRel, name):
        repoBase = os.path.expanduser(repoBase)
        repo = f'{repoBase}/{repoRel}'
        self.repo = repo
        self.version = VERSION
        self.sourceDir = f'{repo}/{SOURCE_DIR}'
        self.imageDir = f'{repo}/{IMAGE_DIR}'
        self._imagery = {}
        self.corpus = f'{repo}/{CORPUS}'
        self.corpusFull = CORPUS_FULL
        TF = Fabric(locations=[self.corpus], modules=[''], silent=True)
        api = TF.load('', silent=True)
        allFeatures = TF.explore(silent=True, show=True)
        loadableFeatures = allFeatures['nodes'] + allFeatures['edges']
        TF.load(loadableFeatures, add=True, silent=True)
        self.api = api
        self._getImagery()
        self.cwd = os.getcwd()
        cwdPat = re.compile(f'^.*/github/([^/]+)/([^/]+)((?:/.+)?)$', re.I)
        cwdRel = cwdPat.findall(self.cwd)
        if cwdRel:
            (thisOrg, thisRepo, thisPath) = cwdRel[0]
            onlineTail = (
                f'{thisOrg}/{thisRepo}'
                f'/blob/master{thisPath}/{name}.ipynb'
            )
        else:
            cwdRel = None
        nbLink = (
            None
            if name is None or cwdRel is None else f'{URL_NB}/{onlineTail}'
        )
        ghLink = (
            None
            if name is None or cwdRel is None else f'{URL_GH}/{onlineTail}'
        )
        docLink = f'https://github.com/{repoRel}/blob/master/docs'
        extraLink = f'https://github.com/Dans-labs/text-fabric/wiki/Cunei'
        dataLink = _outLink(
            self.corpusFull, f'{docLink}/about.md', 'provenance of this corpus'
        )
        featureLink = _outLink(
            'Feature docs', f'{docLink}/transcription.md',
            'feature documentation'
        )
        cuneiLink = _outLink('Cunei API', extraLink, 'cunei api documentation')
        tfLink = _outLink(
            f'Text-Fabric API {api.TF.version}',
            'https://github.com/Dans-labs/text-fabric/wiki/api',
            'text-fabric-api'
        )
        tfsLink = _outLink(
            'Search Reference', (
                'https://github.com/Dans-labs/text-fabric/wiki/api'
                '#search-template-introduction'
            ), 'Search Templates Introduction and Reference'
        )
        dm(
            '**Documentation:**'
            f' {dataLink} {featureLink} {cuneiLink} {tfLink} {tfsLink}'
        )
        if nbLink:
            dm(
                f'''
This notebook online:
{_outLink('NBViewer', nbLink)}
{_outLink('GitHub', ghLink)}
'''
            )
        thisRepoDir = (
            None if cwdRel is None else f'{repoBase}/{thisOrg}/{thisRepo}'
        )
        self.tempDir = (
            None if cwdRel is None else f'{thisRepoDir}/{TEMP_DIR}'
        )
        self.reportDir = (
            None if cwdRel is None else f'{thisRepoDir}/{REPORT_DIR}'
        )
        for cdir in (self.tempDir, self.reportDir):
            if cdir:
                os.makedirs(cdir, exist_ok=True)

        self._loadCSS()

    def getSource(self, node, nodeType=None, lineNumbers=False):
        api = self.api
        F = api.F
        L = api.L
        sourceLines = []
        lineNumber = ''
        if lineNumbers:
            lineNo = F.srcLnNum.v(node)
            lineNumber = f'{lineNo:>5} ' if lineNo else ''
        sourceLine = F.srcLn.v(node)
        if sourceLine:
            sourceLines.append(f'{lineNumber}{sourceLine}')
        for child in L.d(node, nodeType):
            sourceLine = F.srcLn.v(child)
            lineNumber = ''
            if sourceLine:
                if lineNumbers:
                    lineNumber = f'{F.srcLnNum.v(child):>5}: '
                sourceLines.append(f'{lineNumber}{sourceLine}')
        return sourceLines

    def atfFromSign(self, n, flags=False):
        F = self.api.F
        Fs = self.api.Fs
        if F.otype.v(n) != 'sign':
            return '«no sign»'

        grapheme = F.grapheme.v(n)
        if grapheme == '…':
            grapheme = '...'
        primeN = F.prime.v(n)
        prime = ("'" * primeN) if primeN else ''

        variantValue = F.variant.v(n)
        variant = f'~{variantValue}' if variantValue else ''

        modifierValue = F.modifier.v(n)
        modifier = f'@{modifierValue}' if modifierValue else ''
        modifierInnerValue = F.modifierInner.v(n)
        modifierInner = f'@{modifierInnerValue}' if modifierInnerValue else ''

        modifierFirst = F.modifierFirst.v(n)

        repeat = F.repeat.v(n)
        if repeat is None:
            varmod = (
                f'{modifier}{variant}'
                if modifierFirst else f'{variant}{modifier}'
            )
            result = f'{grapheme}{prime}{varmod}'
        else:
            if repeat == -1:
                repeat = 'N'
            varmod = (
                f'{modifierInner}{variant}'
                if modifierFirst else f'{variant}{modifierInner}'
            )
            result = f'{repeat}({grapheme}{prime}{varmod}){modifier}'

        if flags:
            for (flag, char) in FLAGS:
                value = Fs(flag).v(n)
                if value:
                    if type(char) is tuple:
                        result += f'{char[0]}{value}{char[1]}'
                    else:
                        result += char

        return result

    def atfFromQuad(self, n, flags=False, outer=True):
        api = self.api
        E = api.E
        F = api.F
        Fs = api.Fs
        if F.otype.v(n) != 'quad':
            return '«no quad»'

        children = E.sub.f(n)
        if not children or len(children) < 2:
            return f'«quad with less than two sub-quads»'
        result = ''
        for child in children:
            nextChildren = E.op.f(child)
            if nextChildren:
                op = nextChildren[0][1]
            else:
                op = ''
            childType = F.otype.v(child)

            thisResult = (
                self.atfFromQuad(child, flags=flags, outer=False) if
                childType == 'quad' else self.atfFromSign(child, flags=flags)
            )
            result += f'{thisResult}{op}'

        variant = F.variantOuter.v(n)
        variantStr = f'~{variant}' if variant else ''

        flagStr = ''
        if flags:
            for (flag, char) in FLAGS:
                value = Fs(flag).v(n)
                if value:
                    if type(char) is tuple:
                        flagStr += f'{char[0]}{value}{char[1]}'
                    else:
                        flagStr += char

        if variant:
            if flagStr:
                if outer:
                    result = f'|({result}){variantStr}|{flagStr}'
                else:
                    result = f'(({result}){variantStr}){flagStr}'
            else:
                if outer:
                    result = f'|({result}){variantStr}|'
                else:
                    result = f'({result}){variantStr}'
        else:
            if flagStr:
                if outer:
                    result = f'|{result}|{flagStr}'
                else:
                    result = f'({result}){flagStr}'
            else:
                if outer:
                    result = f'|{result}|'
                else:
                    result = f'({result})'

        return result

    def atfFromOuterQuad(self, n, flags=False):
        api = self.api
        F = api.F
        nodeType = F.otype.v(n)
        if nodeType == 'sign':
            return self.atfFromSign(n, flags=flags)
        elif nodeType == 'quad':
            return self.atfFromQuad(n, flags=flags, outer=True)
        else:
            return '«no outer quad»'

    def atfFromCluster(self, n, seen=None):
        api = self.api
        F = api.F
        E = api.E
        if F.otype.v(n) != 'cluster':
            return '«no cluster»'

        typ = F.type.v(n)
        (bOpen, bClose) = CLUSTER_BRACKETS[typ]
        if bClose == ')':
            bClose = ')a'
        children = api.sortNodes(E.sub.f(n))

        if seen is None:
            seen = set()
        result = []
        for child in children:
            if child in seen:
                continue
            childType = F.otype.v(child)

            thisResult = (
                self.atfFromCluster(child, seen=seen) if childType == 'cluster'
                else self.atfFromQuad(child, flags=True) if childType == 'quad'
                else self.atfFromSign(child, flags=True)
                if childType == 'sign' else None
            )
            seen.add(child)
            if thisResult is None:
                dm(
                    f'TF: child of cluster has type {childType}:'
                    ' should not happen'
                )
            result.append(thisResult)
        return f'{bOpen}{" ".join(result)}{bClose}'

    def getOuterQuads(self, n):
        api = self.api
        F = api.F
        E = api.E
        L = api.L
        return [
            quad for quad in L.d(n)
            if (
                F.otype.v(quad) in OUTER_QUAD_TYPES and
                all(F.otype.v(parent) != 'quad' for parent in E.sub.t(quad))
            )
        ]

    def lineFromNode(self, n):
        api = self.api
        F = api.F
        L = api.L
        caseOrLineUp = [m for m in L.u(n) if F.terminal.v(m)]
        return caseOrLineUp[0] if caseOrLineUp else None

    def nodeFromCase(self, passage):
        api = self.api
        F = api.F
        L = api.L
        T = api.T
        section = passage[0:2]
        caseNum = passage[2].replace('.', '')
        column = T.nodeFromSection(section)
        if column is None:
            return None
        casesOrLines = [
            c for c in L.d(column)
            if F.terminal.v(c) and F.number.v(c) == caseNum
        ]
        if not casesOrLines:
            return None
        return casesOrLines[0]

    def caseFromNode(self, n):
        api = self.api
        F = api.F
        T = api.T
        L = api.L
        section = T.sectionFromNode(n)
        if section is None:
            return None
        nodeType = F.otype.v(n)
        if nodeType in {'sign', 'quad', 'cluster', 'case'}:
            if nodeType == 'case':
                caseNumber = F.number.v(n)
            else:
                caseOrLine = [m for m in L.u(n) if F.terminal.v(m)][0]
                caseNumber = F.number.v(caseOrLine)
            return (section[0], section[1], caseNumber)
        else:
            return section

    # this is a slow implementation!
    def _casesByLevelM(self, lev, terminal=True):
        api = self.api
        E = api.E
        F = api.F
        if lev == 0:
            results = F.otype.s('line')
        else:
            parents = self._casesByLevelM(lev - 1, terminal=False)
            results = reduce(
                operator.add,
                [
                    tuple(s for s in E.sub.f(p) if F.otype.v(s) == 'case')
                    for p in parents
                ],
                (),
            )
        return (
            tuple(r for r in results if F.terminal.v(r))
            if terminal else results
        )

    # this is a fast implementation!
    def _casesByLevelS(self, lev, terminal=True):
        api = self.api
        S = api.S
        sortNodes = api.sortNodes
        query = ''
        for i in range(lev + 1):
            extra = (' terminal' if i == lev and terminal else '')
            nodeType = 'line' if i == 0 else 'case'
            query += ('  ' * i) + f'w{i}:{nodeType}{extra}\n'
        for i in range(lev):
            query += f'w{i} -sub> w{i+1}\n'
        results = list(S.search(query))
        return sortNodes(tuple(r[-1] for r in results))

    def casesByLevel(self, lev, terminal=True):
        api = self.api
        F = api.F
        return ((
            tuple(c for c in F.otype.s('line') if F.terminal.v(c))
            if terminal else F.otype.s('line')
        ) if lev == 0 else (
            tuple(c for c in F.depth.s(lev) if F.terminal.v(c))
            if terminal else F.depth.s(lev)
        ))

    def lineart(self, ns, key=None, asLink=False, withCaption=None, **options):
        return self._getImages(
            ns,
            kind='lineart',
            key=key,
            asLink=asLink,
            withCaption=withCaption,
            **options
        )

    def photo(self, ns, key=None, asLink=False, withCaption=None, **options):
        return self._getImages(
            ns,
            kind='photo',
            key=key,
            asLink=asLink,
            withCaption=withCaption,
            **options
        )

    def imagery(self, objectType, kind):
        return set(self._imagery.get(objectType, {}).get(kind, {}))

    def cdli(self, n, linkText=None, asString=False):
        (nType, objectType, identifier) = self._imageClass(n)
        if linkText is None:
            linkText = identifier
        result = _wrapLink(linkText, objectType, 'main', identifier)
        if asString:
            return result
        else:
            display(HTML(result))

    def tabletLink(self, t, text=None, asString=False):
        api = self.api
        L = api.L
        F = api.F
        if type(t) is str:
            pNum = t
        else:
            n = t if F.otype.v(t) == 'tablet' else L.u(t, otype='tablet')[0]
            pNum = F.catalogId.v(n)

        title = ('to CDLI main page for this tablet')
        linkText = pNum if text is None else text
        url = URL_FORMAT['tablet']['main'].format(pNum)

        result = _outLink(linkText, url, title=title)
        if asString:
            return result
        display(HTML(result))

    def plain(
        self,
        n,
        linked=True,
        lineart=True,
        withNodes=False,
        lineNumbers=False,
        asString=False,
    ):
        api = self.api
        F = api.F

        nType = F.otype.v(n)
        markdown = ''
        nodeRep = f' *{n}* ' if withNodes else ''

        if nType in ATF_TYPES:
            isSign = nType == 'sign'
            isQuad = nType == 'quad'
            rep = (
                self.atfFromSign(n) if isSign else self.atfFromQuad(n)
                if isQuad else self.atfFromCluster(n)
            )
            if linked:
                rep = self.tabletLink(n, text=rep, asString=True)
            theLineart = ''
            if lineart:
                if isSign or isQuad:
                    width = '2em' if isSign else '4em'
                    height = '4em' if isSign else '6em'
                    theLineart = self._getImages(
                        n,
                        kind='lineart',
                        width=width,
                        height=height,
                        asString=True,
                        withCaption=False,
                        warning=False
                    )
                    theLineart = f' {theLineart}'
            markdown = (
                f'{rep}{nodeRep}{theLineart}'
            ) if theLineart else f'{rep}{nodeRep}'
        elif nType == 'comment':
            rep = F.type.v(n)
            if linked:
                rep = self.tabletLink(n, text=rep, asString=True)
            markdown = f'{rep}{nodeRep}: {F.text.v(n)}'
        elif nType == 'line' or nType == 'case':
            rep = f'{nType} {F.number.v(n)}'
            if linked:
                rep = self.tabletLink(n, text=rep, asString=True)
            theLine = ''
            if lineNumbers:
                if F.terminal.v(n):
                    theLine = f' @{F.srcLnNum.v(n)} '
            markdown = f'{rep}{nodeRep}{theLine}'
        elif nType == 'column':
            rep = f'{nType} {F.number.v(n)}'
            if linked:
                rep = self.tabletLink(n, text=rep, asString=True)
            theLine = ''
            if lineNumbers:
                theLine = f' @{F.srcLnNum.v(n)} '
            markdown = f'{rep}{nodeRep}{theLine}'
        elif nType == 'face':
            rep = f'{nType} {F.type.v(n)}'
            if linked:
                rep = self.tabletLink(n, text=rep, asString=True)
            theLine = ''
            if lineNumbers:
                theLine = f' @{F.srcLnNum.v(n)} '
            markdown = f'{rep}{nodeRep}{theLine}'
        elif nType == 'tablet':
            rep = f'{nType} {F.catalogId.v(n)}'
            if linked:
                rep = self.tabletLink(n, text=rep, asString=True)
            theLine = ''
            if lineNumbers:
                theLine = f' @{F.srcLnNum.v(n)} '
            markdown = f'{rep}{nodeRep}{theLine}'

        if asString:
            return markdown
        dm(markdown)

    def plainTuple(
        self,
        ns,
        seqNumber,
        linked=1,
        lineart=True,
        withNodes=False,
        lineNumbers=False,
        asString=False,
    ):
        markdown = [str(seqNumber)]
        for (i, n) in enumerate(ns):
            markdown.append(
                self.plain(
                    n,
                    linked=i == linked - 1,
                    lineart=lineart,
                    withNodes=withNodes,
                    lineNumbers=lineNumbers,
                    asString=True,
                )
            )
        markdown = '|'.join(markdown)
        if asString:
            return markdown
        api = self.api
        F = api.F
        head = ['n | ' + (' | '.join(F.otype.v(n) for n in ns))]
        head.append(' | '.join('---' for n in range(len(ns) + 1)))
        head.append(markdown)

        dm('\n'.join(head))

    def pretty(
        self,
        n,
        lineart=True,
        withNodes=False,
        lineNumbers=False,
        suppress=set(),
        highlights=set()
    ):
        html = []
        self._pretty(
            n,
            True,
            html,
            seen=set(),
            lineart=lineart,
            withNodes=withNodes,
            lineNumbers=lineNumbers,
            suppress=suppress,
            highlights=highlights,
        )
        htmlStr = '\n'.join(html)
        display(HTML(htmlStr))

    def prettyTuple(
        self,
        ns,
        seqNumber,
        item='Result',
        lineart=True,
        withNodes=False,
        lineNumbers=False,
        suppress=set(),
    ):
        api = self.api
        L = api.L
        F = api.F
        sortNodes = api.sortNodes
        tablets = set()
        highlights = set()
        for n in ns:
            nType = F.otype.v(n)
            if nType == 'tablet':
                tablets.add(n)
            else:
                t = L.u(n, otype='tablet')[0]
                tablets.add(t)
                highlights.add(n)
        dm(f'''
##### {item} {seqNumber}
''')
        for t in sortNodes(tablets):
            self.pretty(
                t,
                lineart=lineart,
                withNodes=withNodes,
                lineNumbers=lineNumbers,
                suppress=suppress,
                highlights=highlights,
            )

    def search(self, query):
        api = self.api
        S = api.S
        results = list(S.search(query))
        nResults = len(results)
        plural = '' if nResults == 1 else 's'
        print(f'{nResults} result{plural}')
        return results

    def table(
        self,
        results,
        start=None,
        end=None,
        linked=1,
        lineart=True,
        withNodes=False,
        lineNumbers=False,
        asString=False,
    ):
        api = self.api
        F = api.F

        collected = []
        if start is None:
            start = 1
        i = -1
        rest = 0
        if not hasattr(results, 'len'):
            if end is None or end > LIMIT_TABLE:
                end = LIMIT_TABLE
            for result in results:
                i += 1
                if i < start - 1:
                    continue
                if i >= end:
                    break
                collected.append((i + 1, result))
        else:
            if end is None:
                end = len(results)
            rest = 0
            if end - (start - 1) > LIMIT_TABLE:
                rest = end - (start - 1) - LIMIT_TABLE
                end = start - 1 + LIMIT_TABLE
            for i in range(start - 1, end):
                collected.append((i + 1, results[i]))

        if len(collected) == 0:
            return
        (firstSeq, firstResult) = collected[0]
        nColumns = len(firstResult)
        markdown = ['n | ' + (' | '.join(F.otype.v(n) for n in firstResult))]
        markdown.append(' | '.join('---' for n in range(nColumns + 1)))
        for (seqNumber, ns) in collected:
            markdown.append(
                self.plainTuple(
                    ns,
                    seqNumber,
                    linked=linked,
                    lineart=lineart,
                    withNodes=withNodes,
                    lineNumbers=lineNumbers,
                    asString=True,
                )
            )
        markdown = '\n'.join(markdown)
        if asString:
            return markdown
        dm(markdown)
        if rest:
            dm(
                f'**{rest} more results skipped**'
                f' because we show a maximum of'
                f' {LIMIT_TABLE} results at a time'
            )

    def show(
        self,
        results,
        condensed=True,
        start=None,
        end=None,
        lineart=True,
        withNodes=False,
        lineNumbers=False,
        suppress=set()
    ):
        if condensed:
            results = self._condense(results)
        if start is None:
            start = 1
        i = -1
        if not hasattr(results, 'len'):
            if end is None or end > LIMIT_SHOW:
                end = LIMIT_SHOW
            for result in results:
                i += 1
                if i < start - 1:
                    continue
                if i >= end:
                    break
                self.prettyTuple(
                    result,
                    i + 1,
                    item='Tablet' if condensed else 'Result',
                    lineart=lineart,
                    withNodes=withNodes,
                    lineNumbers=lineNumbers,
                    suppress=suppress
                )
        else:
            if end is None:
                end = len(results)
            rest = 0
            if end - (start - 1) > LIMIT_SHOW:
                rest = end - (start - 1) - LIMIT_SHOW
                end = start - 1 + LIMIT_SHOW
            for i in range(start - 1, end):
                self.prettyTuple(
                    results[i],
                    i + 1,
                    item='Tablet' if condensed else 'Result',
                    lineart=lineart,
                    withNodes=withNodes,
                    lineNumbers=lineNumbers,
                    suppress=suppress
                )
            if rest:
                dm(
                    f'**{rest} more results skipped**'
                    f' because we show a maximum of'
                    f' {LIMIT_SHOW} results at a time'
                )

    def _condense(self, results):
        api = self.api
        F = api.F
        L = api.L
        sortNodes = api.sortNodes
        tablets = {}
        for ns in results:
            for n in ns:
                if F.otype.v(n) == 'tablet':
                    tablets.setdefault(n, set())
                else:
                    t = L.u(n, otype='tablet')[0]
                    tablets.setdefault(t, set()).add(n)
        return tuple((t, ) + tuple(tablets[t]) for t in sortNodes(tablets))

    def _pretty(
        self,
        n,
        outer,
        html,
        lineart=True,
        withNodes=False,
        lineNumbers=False,
        seen=set(),
        suppress=set(),
        highlights=set(),
    ):
        api = self.api
        L = api.L
        F = api.F
        E = api.E
        sortNodes = api.sortNodes
        hl = ' hl' if n in highlights else ''
        nType = F.otype.v(n)
        className = nType
        nodePart = (f'<span class="nd">{n}</span>' if withNodes else '')
        heading = ''
        featurePart = ''
        commentsPart = self._getComments(
            n, withNodes=withNodes, lineNumbers=lineNumbers
        ) if nType in COMMENT_TYPES else ''
        children = ()

        if nType == 'tablet':
            heading = F.catalogId.v(n)
            heading += ' '
            heading += self._getFeatures(
                n,
                suppress,
                ('name', 'period', 'excavation'),
                plain=True,
            )
            children = L.d(n, otype='face')
        elif nType == 'face':
            heading = F.type.v(n)
            featurePart = self._getFeatures(
                n,
                suppress,
                ('identifier', 'fragment'),
            )
            children = L.d(n, otype='column')
        elif nType == 'column':
            heading = F.number.v(n)
            if F.prime.v(n):
                heading += "'"
            children = L.d(n, otype='line')
        elif nType == 'line' or nType == 'case':
            heading = F.number.v(n)
            if F.prime.v(n):
                heading += "'"
            if F.terminal.v(n):
                className = 'trminal'
                theseFeats = ('srcLnNum', ) if lineNumbers else ()
                featurePart = self._getFeatures(
                    n,
                    suppress,
                    theseFeats,
                )
                children = sortNodes(
                    set(L.d(n, otype='cluster'))
                    | set(L.d(n, otype='quad'))
                    | set(L.d(n, otype='sign'))
                )
            else:
                children = E.sub.f(n)
        elif nType == 'comment':
            heading = F.type.v(n)
            featurePart = self._getFeatures(
                n,
                suppress,
                ('text', ),
            )
        elif nType == 'cluster':
            seen.add(n)
            heading = CLUSTER_TYPES.get(F.type.v(n), '')
            children = sortNodes(
                set(L.d(n, otype='cluster'))
                | set(L.d(n, otype='quad'))
                | set(L.d(n, otype='sign'))
            )
        elif nType == 'quad':
            seen.add(n)
            children = E.sub.f(n)
        elif nType == 'sign':
            featurePart = self._getAtf(n)
            seen.add(n)
            if not outer and F.type.v(n) == 'empty':
                return

        if outer:
            typePart = self.tabletLink(
                n, text=f'{nType} {heading}', asString=True
            )
        else:
            typePart = heading

        isCluster = nType == 'cluster'
        extra = 'B' if isCluster else ''
        label = f'''
    <div class="lbl {className}{extra}">
        {typePart}
        {nodePart}
    </div>
''' if typePart or nodePart else ''

        if isCluster:
            if outer:
                html.append(f'<div class="contnr {className} {hl}">')
            html.append(label)
            if outer:
                html.append(f'<div class="children {className}">')
        else:
            html.append(
                f'''
<div class="contnr {className} {hl}">
    {label}
    <div class="meta">
        {featurePart}
        {commentsPart}
    </div>
'''
            )
        if lineart:
            isQuad = nType == 'quad'
            isSign = nType == 'sign'
            if isQuad or isSign:
                isOuter = outer or (
                    all(F.otype.v(parent) != 'quad' for parent in E.sub.t(n))
                )
                if isOuter:
                    width = '2em' if isSign else '4em'
                    height = '4em' if isSign else '6em'
                    theLineart = self._getImages(
                        n,
                        kind='lineart',
                        width=width,
                        height=height,
                        asString=True,
                        withCaption=False,
                        warning=False
                    )
                    if theLineart:
                        html.append(f'<div>{theLineart}</div>')
        caseDir = ''
        if not isCluster:
            if children:
                if nType == 'case':
                    depth = F.depth.v(n)
                    caseDir = 'v' if depth & 1 else 'h'
                html.append(
                    f'''
    <div class="children {className}{caseDir}">
'''
                )

        for ch in children:
            if ch not in seen:
                self._pretty(
                    ch,
                    False,
                    html,
                    seen=seen,
                    lineart=lineart,
                    withNodes=withNodes,
                    lineNumbers=lineNumbers,
                    suppress=suppress,
                    highlights=highlights,
                )
                if nType == 'quad':
                    nextChildren = E.op.f(ch)
                    if nextChildren:
                        op = nextChildren[0][1]
                        html.append(f'<div class="op">{op}</div>')
        if isCluster:
            html.append(
                f'''
    <div class="lbl {className}E {hl}">
        {typePart}
        {nodePart}
    </div>
'''
            )
            if outer:
                html.append('</div></div>')
        else:
            if children:
                html.append('''
    </div>
''')
            html.append('''
</div>
''')

    def _getFeatures(self, n, suppress, features, plain=False):
        api = self.api
        Fs = api.Fs
        featurePart = '' if plain else '<div class="features">'
        for name in features:
            if 'name' not in suppress:
                value = Fs(name).v(n)
                if value is not None:
                    featurePart += (
                        f' <span class="{name}">{Fs(name).v(n)}</span>'
                    )
        if not plain:
            featurePart += '</div>'
        return featurePart

    def _getComments(self, n, withNodes=False, lineNumbers=False):
        api = self.api
        E = api.E
        cns = E.comments.f(n)
        if len(cns):
            html = ['<div class="comments">']
            for c in cns:
                self._pretty(
                    c,
                    False,
                    html,
                    lineart=False,
                    withNodes=withNodes,
                    lineNumbers=lineNumbers,
                )
            html.append('</div>')
            commentsPart = ''.join(html)
        else:
            commentsPart = ''
        return commentsPart

    def _getAtf(self, n):
        atf = self.atfFromSign(n, flags=True)
        featurePart = f' <span class="srcLn">{atf}</span>'
        return featurePart

    def _loadCSS(self):
        display(HTML(CSS))

    def _imageClass(self, n):
        api = self.api
        F = api.F
        if type(n) is str:
            identifier = n
            if n == '':
                identifier = None
                objectType = None
                nType = None
            elif len(n) == 1:
                objectType = 'ideograph'
                nType = 'sign/quad'
            else:
                if n[0] == 'P' and n[1:].isdigit():
                    objectType = 'tablet'
                    nType = 'tablet'
                else:
                    objectType = 'ideograph'
                    nType = 'sign/quad'
        else:
            nType = F.otype.v(n)
            if nType in OUTER_QUAD_TYPES:
                identifier = self.atfFromOuterQuad(n)
                objectType = 'ideograph'
            elif nType == 'tablet':
                identifier = F.catalogId.v(n)
                objectType = 'tablet'
            else:
                identifier = None
                objectType = None
        return (nType, objectType, identifier)

    def _getImages(
        self,
        ns,
        kind=None,
        key=None,
        asLink=False,
        withCaption=None,
        warning=True,
        asString=False,
        **options
    ):
        if type(ns) is int or type(ns) is str:
            ns = [ns]
        result = []
        attStr = ' '.join(
            f'{opt}="{value}"' for (opt, value) in options.items()
            if opt not in SIZING
        )
        cssProps = {}
        for (opt, value) in options.items():
            if opt in SIZING:
                if type(value) is int:
                    force = False
                    realValue = f'{value}px'
                else:
                    if value.startswith('!'):
                        force = True
                        realValue = value[1:]
                    else:
                        force = False
                        realValue = value
                    if realValue.isdecimal():
                        realValue += 'px'
                cssProps[f'max-{opt}'] = realValue
                if force:
                    cssProps[f'min-{opt}'] = realValue
        cssStr = ' '.join(
            f'{opt}: {value};' for (opt, value) in cssProps.items()
        )
        if withCaption is None:
            withCaption = None if asLink else 'bottom'
        for n in ns:
            caption = None
            (nType, objectType, identifier) = self._imageClass(n)
            if objectType:
                imageBase = self._imagery.get(objectType, {}).get(kind, {})
                images = imageBase.get(identifier, None)
                if withCaption:
                    caption = _wrapLink(
                        identifier, objectType, 'main', identifier
                    )
                if images is None:
                    thisImage = (
                        f'<span><b>no {kind}</b> for {objectType}'
                        f' <code>{identifier}</code></span>'
                    ) if warning else ''
                else:
                    image = images.get(key or '', None)
                    if image is None:
                        thisImage = (
                            '<span><b>try</b>'
                            ' key=<code>'
                            f'{"</code> <code>".join(sorted(images.keys()))}'
                            '</code></span>'
                        ) if warning else ''
                    else:
                        if asLink:
                            thisImage = identifier
                        else:
                            theImage = self._useImage(
                                image, kind, key or '', n
                            )
                            thisImage = (
                                f'<img src="{theImage}"'
                                f' style="display: inline;{cssStr}"'
                                f' {attStr} />'
                            )
                thisResult = _wrapLink(
                    thisImage,
                    objectType,
                    kind,
                    identifier,
                    pos=withCaption,
                    caption=caption
                ) if thisImage else None
            else:
                thisResult = (
                    f'<span><b>no {kind}</b> for'
                    f' <code>{nType}</code>s</span>'
                ) if warning else ''
            result.append(thisResult)
        if not warning:
            result = [image for image in result if image]
        if not result:
            return ''
        if asString:
            return ''.join(result)
        resultStr = f'</div><div style="{ITEM_STYLE}">'.join(result)
        html = (
            f'<div style="{FLEX_STYLE}">'
            f'<div style="{ITEM_STYLE}">'
            f'{resultStr}</div></div>'
        ).replace('\n', '')
        display(HTML(html))
        if not warning:
            return True

    def _useImage(self, image, kind, key, node):
        api = self.api
        F = api.F
        (imageDir, imageName) = os.path.split(image)
        (base, ext) = os.path.splitext(imageName)
        localDir = f'{self.cwd}/{LOCAL_DIR}'
        if not os.path.exists(localDir):
            os.makedirs(localDir, exist_ok=True)
        if type(node) is int:
            nType = F.otype.v(node)
            if nType == 'tablet':
                nodeRep = F.catalogId.v(node)
            elif nType in OUTER_QUAD_TYPES:
                nodeRep = self.atfFromOuterQuad(node)
            else:
                nodeRep = str(node)
        else:
            nodeRep = node
        nodeRep = (
            nodeRep.lower().replace('|', 'q').replace('~', '-')
            .replace('@', '(a)').replace('&', '(e)')
            .replace('+', '(p)').replace('.', '(d)')
        )
        keyRep = '' if key == '' else f'-{key}'
        localImageName = f'{kind}-{nodeRep}{keyRep}{ext}'
        localImagePath = f'{localDir}/{localImageName}'
        if (
            not os.path.exists(localImagePath) or
            os.path.getmtime(image) > os.path.getmtime(localImagePath)
        ):
            copyfile(image, localImagePath)
        return f'{LOCAL_DIR}/{localImageName}'

    def _getImagery(self):
        for (dirFmt, ext, kind, objectType) in (
            (IDEO_TO, LINEART_EXT, 'lineart', 'ideograph'),
            (TABLET_TO, LINEART_EXT, 'lineart', 'tablet'),
            (PHOTO_TO, PHOTO_EXT, 'photo', 'tablet'),
        ):
            srcDir = dirFmt.format(self.imageDir)
            filePaths = glob(f'{srcDir}/*.{ext}')
            images = {}
            idPat = re.compile('P[0-9]+')
            for filePath in filePaths:
                (fileDir, fileName) = os.path.split(filePath)
                (base, thisExt) = os.path.splitext(fileName)
                if kind == 'lineart' and objectType == 'tablet':
                    ids = idPat.findall(base)
                    if not ids:
                        print(f'skipped non-{objectType} "{fileName}"')
                        continue
                    identifier = ids[0]
                    key = base.replace('_l', '').replace(identifier, '')
                else:
                    identifier = base
                    key = ''
                images.setdefault(identifier, {})[key] = filePath
            self._imagery.setdefault(objectType, {})[kind] = images
            print(f'Found {len(images)} {objectType} {kind}s')



import sys
import time


class Timestamp(object):
    def __init__(self, level=None):
        self.oneLevelRep = '   |   '
        self.timestamp = {}
        self.indent(level=level, reset=True)
        self.log = []
        self.verbose = -2

    def raw_msg(self, msg, tm=True, nl=True, cache=0, error=False):
        # cache = -1: only to cache
        # cache =  1: to cache and to console
        # cache =  0: only to console
        if self.verbose != -2 and self.level >= self.verbose:
            return
        if tm:
            msgRep = '{}{:>7} {}'.format(self.levelRep, self._elapsed(), msg
                                         ).replace('\n', '\n' + self.levelRep)
        else:
            msgRep = '{}{}'.format(self.levelRep,
                                   msg).replace('\n', '\n' + self.levelRep)
        if cache:
            self.log.append((error, nl, msgRep))
        if cache >= 0:
            channel = sys.stderr if error else sys.stdout
            channel.write('{}{}'.format(msgRep, '\n' if nl else ''))
            channel.flush()

    def cache(self):
        for (error, nl, msgRep) in self.log:
            channel = sys.stderr if error else sys.stdout
            channel.write('{}{}'.format(msgRep, '\n' if nl else ''))
        sys.stderr.flush()
        sys.stdout.flush()
        self.log = []

    def info(self, msg, tm=True, nl=True, cache=False):
        self.raw_msg(msg, tm=tm, nl=nl, cache=cache)

    def error(self, msg, tm=True, nl=True, cache=False):
        self.raw_msg(msg, tm=tm, nl=nl, cache=cache, error=True)

    def indent(self, level=None, reset=False, verbose=None):
        self.level = level if level is not None else 0
        self.levelRep = self.oneLevelRep * self.level
        if reset:
            self.timestamp[self.level] = time.time()
        if verbose is not None:
            self.verbose = verbose

    def _elapsed(self):
        interval = time.time() - self.timestamp.setdefault(
            self.level, time.time()
        )
        if interval < 10:
            return "{: 2.2f}s".format(interval)
        interval = int(round(interval))
        if interval < 60:
            return "{:>2d}s".format(interval)
        if interval < 3600:
            return "{:>2d}m {:>02d}s".format(interval // 60, interval % 60)
        return "{:>2d}h {:>02d}m {:>02d}s".format(
            interval // 3600, (interval % 3600) // 60, interval % 60
        )
